﻿/*	Application		IPC receiver-side using socket (UDP) - Message serialisation example		
	File			Program.cs (IPC_Receiver_MessageSerialisation)
	Purpose			Demonstration of local IPC, with a serialised data structure, using a blocking socket (on the receiving side)
	Author			Richard John Anthony
	Date			January 2015
	
	Function		1. Creates a socket, and binds to local address (uses INADDR_ANY)
					2. Waits for a message to arrive on its port (8007)
					3. Displays the message 
					4. Exits once a key is pressed
*/

using System;
using System.Text;
using System.Net;
using System.Net.Sockets;
using System.Runtime.InteropServices;

namespace IPC_Receiver_MessageSerialisation
{
    class C_IPC_socket_Receiver
    {
        public const Int32 RECEIVE_PORT = 8007;
        Socket m_IPC_ReceiveSOCKET;
        public byte[] m_bRecdArray;
        IPEndPoint m_localIPEndPoint;   // Local IP address and port
        IPEndPoint m_SenderEndPoint;    // This will hold the address and port of the sender after a message has been received

        public const Int32 MAX_NAME_LENGTH = 20;
        public const Int32 MAX_ADDRESS_LENGTH = 50;

        // Message structure (this will be deserialised from the received message before displaying contents)
        [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Ansi, Pack = 1)]
        public unsafe struct MyDetailsMessage
        {
            [MarshalAs(UnmanagedType.ByValTStr, SizeConst = MAX_NAME_LENGTH)]
            public byte[] cName;    // (c++)  char cName[MAX_NAME_LENGTH];

	        public Int32 iAge;

            [MarshalAs(UnmanagedType.ByValTStr, SizeConst = MAX_ADDRESS_LENGTH)]
            public byte[] cAddress; // (c++)  char cAddress[MAX_ADDRESS_LENGTH];
        };
        
        static void Main(string[] args)
        {
            Console.WriteLine("          Local Inter-Process Communication (receiver side).");
            Console.WriteLine("  This program waits for a message to arrive on port 8007 (decimal).");
            Console.WriteLine("The message is displayed and the process exits when a key is pressed.");
            Console.WriteLine("---------------------------------------------------------------------");
            Console.WriteLine("Waiting for message ...   ");

            C_IPC_socket_Receiver IPC_socket_Receiver = new C_IPC_socket_Receiver();
        }

        C_IPC_socket_Receiver()
        {
            bool bSuccess;
            
            // Create receive socket - this is the logical 'communication endpoint'
            bSuccess = CreateReceiveSocket();
            if (false == bSuccess)
            {
                Console.WriteLine("Exiting the program because a socket could not be created");
                CloseSocketAndExit();
            }

            // Set the Local Address details (IP address and  port number)
            SetUpLocalAddressStruct();

            // Bind (The operating system makes a mapping between this process and the requested Port)
            bSuccess = BindToLocalAddress();
            if (false == bSuccess)
            {
                Console.WriteLine("Exiting the program because binding to the requested port was not allowed (port may be in use)");
                CloseSocketAndExit();
            }

            // Wait for, and receive the message
            m_bRecdArray = ReadMessageFromReceiveBufferOrWait();
            MyDetailsMessage RecdMessage = DeSerialize_IPC_Reply(m_bRecdArray);
 
            // The data values are read from the MyDetailsMessage structure using the specific field names
            // (which are cName, cAddress and iAge).
            Console.WriteLine("   Name: {0}\n   Address: {1}\n   Age: {2} ",
                ByteArrayToString(RecdMessage.cName), ByteArrayToString(RecdMessage.cAddress), RecdMessage.iAge);
       
            CloseSocketAndExit();
        }

        bool CreateReceiveSocket()
        {
            try
            {
                m_IPC_ReceiveSOCKET = new Socket(AddressFamily.InterNetwork, SocketType.Dgram, ProtocolType.Udp);
            }
            catch (SocketException)  // Catch exceptions from the socket operations
            {
                Console.WriteLine("CreateSocket() failed");
                return false;
            }
            return true;
        }

        void SetUpLocalAddressStruct()
        {   // Create an Endpoint that will cause the listening activity to apply to all the local node's interfaces
            m_localIPEndPoint = new System.Net.IPEndPoint(IPAddress.Any, System.Convert.ToInt16(RECEIVE_PORT));
        }

        bool BindToLocalAddress()
        {
            try
            {   // Bind to the local IP Address and selected port
                m_IPC_ReceiveSOCKET.Bind(m_localIPEndPoint);
                return true;
            }
            catch   // Catch any exceptions
            {
                Console.WriteLine("bind() Failed!");
                return false;
            }
        }

        byte[] ReadMessageFromReceiveBufferOrWait()
        {
            byte[] bArray = new byte[74];
            int iReceiveByteCount;
            try
            {

                m_SenderEndPoint = new IPEndPoint(IPAddress.Any, 0);   // Initialise (will be overwritten)
                EndPoint SenderAddress = (EndPoint)m_SenderEndPoint;
                iReceiveByteCount = m_IPC_ReceiveSOCKET.ReceiveFrom(bArray, ref SenderAddress);

                // Place the sender's address into the m_SenderEndPoint, from the SenderAddress EndPoint
                m_SenderEndPoint.Address = IPAddress.Parse(((IPEndPoint)SenderAddress).Address.ToString());
            }
            catch (SocketException ex)
            {
                Console.WriteLine("Receive failed");
                CloseSocketAndExit();
            }

            Console.WriteLine("Message received from sender at address: {0}", m_SenderEndPoint.Address);
            return bArray;
        }

        public MyDetailsMessage DeSerialize_IPC_Reply(byte[] bRecdByteArray)
        {   // Converts a byte array (from Receive) to a MyDetailsMessage structure (Fields have fixed length, but non-fixed data length)
            MyDetailsMessage result_MyDetailsMessage;
            int iFieldSize1 = MAX_NAME_LENGTH;      // The DataLength (returned by MyDetailsMessage.cName.Length) is NOT the same as the FieldSize
            result_MyDetailsMessage.iAge = 0;       // Field must be initialised before SizeOf() used
            int iFieldSize2 = Marshal.SizeOf(result_MyDetailsMessage.iAge);
            int iFieldSize3 = MAX_ADDRESS_LENGTH;   // The DataLength (returned by MyDetailsMessage.cAddress.Length) is NOT the same as the FieldSize
            int TotalSize = iFieldSize1 + iFieldSize2 + iFieldSize3;

            int iFieldOffset = 0;           // Create an offset value which ensures the PDU fields are constructed from the correct parts of the array 
            result_MyDetailsMessage.cName = new byte[MAX_NAME_LENGTH];
            for (int iIndex = 0; iIndex < iFieldSize1; iIndex++)
            {   // Byte-by-byte copy data from the byte array into the data field  
                {
                    result_MyDetailsMessage.cName[iIndex] = bRecdByteArray[iIndex + iFieldOffset];
                }
            }

   
            iFieldOffset += iFieldSize1;
            result_MyDetailsMessage.iAge = ByteArrayToInt32(bRecdByteArray, iFieldOffset);

            iFieldOffset += iFieldSize2;
            result_MyDetailsMessage.cAddress = new byte[MAX_ADDRESS_LENGTH];
            for (int iIndex = 0; iIndex < iFieldSize3; iIndex++)
            {   // Byte-by-byte copy data from the byte array into the data field  
                {
                    result_MyDetailsMessage.cAddress[iIndex] = bRecdByteArray[iIndex + iFieldOffset];
                }
            }
            return result_MyDetailsMessage;
        }

        void CloseSocketAndExit()
        {
            m_IPC_ReceiveSOCKET.Close();
            Console.WriteLine("Socket closed");
            Console.WriteLine("(press a key to continue)");
            Console.ReadKey();	// Wait for user to type a key (to allow them to read the message before the window is closed)
            System.Environment.Exit(1);
        }

        static byte[] StringToByteArray(string sStr)
        {
            byte[] bArray = new byte[sStr.Length * sizeof(char)];
            System.Buffer.BlockCopy(sStr.ToCharArray(), 0, bArray, 0, bArray.Length);
            return bArray;
        }

        static string ByteArrayToString(byte[] bArray)
        {
            char[] cArray = new char[bArray.Length / sizeof(char)];
            System.Buffer.BlockCopy(bArray, 0, cArray, 0, bArray.Length);
            return new string(cArray);
        }

        static byte[] Int32ToByteArray(Int32 iInt32)
        {   // The int parameter is treated as Int32 by GetBytes() so the resulting array contains 4 bytes
            byte[] bArray = BitConverter.GetBytes(iInt32);
            if (BitConverter.IsLittleEndian)
            {   // Make it platform independent
                Array.Reverse(bArray);
            }
            return bArray;
        }

        static Int32 ByteArrayToInt32(byte[] bArray, int iIndex)
        {   // Int32 contains 4 bytes. ToInt32() converts 4 bytes starting from iIndex to Int32 
            return (Int32) BitConverter.ToInt32(bArray, iIndex);
        }
    }
}