﻿/*	Application		IPC sender-side using socket (UDP) - Message serialisation example		
	File			Program.cs (IPC_Sender_MessageSerialisation)
	Purpose			Demonstration of local IPC, with a serialised data structure, using a blocking socket (on the receiving side)
	Author			Richard John Anthony
	Date			January 2015
	
	Function		1. Creates a socket
					2. Waits for user to type a message
					3. Sends the message to the the application port (8007) on the local computer (uses local loopback address 127.0.0.1)
					4. Exits once the message has been sent
*/

using System;
using System.Text;
using System.Net;
using System.Net.Sockets;
using System.Runtime.InteropServices;

namespace IPC_socket_Sender_with_Reply
{
    class C_IPC_socket_Sender
    {
        public const Int32 SEND_PORT = 8007;
        Socket m_IPC_SendSOCKET;
        IPEndPoint m_SendEndPoint;      // The address and port to send to

        public const Int32 MAX_NAME_LENGTH = 20;
        public const Int32 MAX_ADDRESS_LENGTH = 50;

        // Message structure (this will be serialised before sending)
        [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Ansi, Pack = 1)]
        public unsafe struct MyDetailsMessage
        {
            [MarshalAs(UnmanagedType.ByValTStr, SizeConst = MAX_NAME_LENGTH)]
            public byte[] cName;    // (c++)  char cName[MAX_NAME_LENGTH];

	        public int iAge;

            [MarshalAs(UnmanagedType.ByValTStr, SizeConst = MAX_ADDRESS_LENGTH)]
            public byte[] cAddress; // (c++)  char cAddress[MAX_ADDRESS_LENGTH];
        };
        
        static void Main(string[] args)
        {
            Console.WriteLine("Local Inter-Process Communication (sender side)");
            Console.WriteLine("This program waits for the user to type a message,");
            Console.WriteLine("and sends it to port 8007 on the local computer.");
            Console.WriteLine("------------------------------------------------------");
            Console.WriteLine("Type a  message to send ...   ");

            C_IPC_socket_Sender IPC_socket_Sender = new C_IPC_socket_Sender();
        }

        C_IPC_socket_Sender()
        {
            bool bSuccess;
            string sIn;
            MyDetailsMessage Message_PDU;
            // The message to be sent will be held in the form of a MyDetailsMessage structure.
            // The data values are written into the structure using the specific field names
            // (which are cName, cAddress and iAge).
            Console.WriteLine("Customer name:		");
            sIn = Console.ReadLine();
            Message_PDU.cName = StringToByteArray(sIn);

            Console.WriteLine("Customer address:	");
            sIn = Console.ReadLine();
            Message_PDU.cAddress = StringToByteArray(sIn);

            Console.WriteLine("Customer age:		");
            sIn = Console.ReadLine();
            try
            {
                Message_PDU.iAge = Convert.ToInt32(sIn);
            }
            catch (FormatException ex)
            {
                Message_PDU.iAge = 0;
                Console.WriteLine("!! Invalid value (Customer age) !!");
            }
            
            // Create Send socket
            bSuccess = CreateSendSocket();
            if (false == bSuccess)
            {
                Console.WriteLine("Exiting because a socket could not be created");
                CloseSocketAndExit();
            }

            // Set the Send Address details (Use local loopback IP address (127.0.0.1), and port 8007)
            SetUpSendAddressStructFor_LocalLoopback();

            // Send the message that the user typed, to the sepcified port on the local computer
            SendMessage(Message_PDU);

            CloseSocketAndExit();
        }

        bool CreateSendSocket()
        {
            try
            {
                m_IPC_SendSOCKET = new Socket(AddressFamily.InterNetwork, SocketType.Dgram, ProtocolType.Udp);
            }
            catch (SocketException)  // Catch exceptions from the socket operations
            {
                Console.WriteLine("CreateSocket() failed");
                return false;
            }
            return true;
        }

        void SetUpSendAddressStructFor_LocalLoopback()
        {
            m_SendEndPoint = new System.Net.IPEndPoint(IPAddress.Loopback, System.Convert.ToInt16(SEND_PORT));
        }

        private byte[] Serialize_Message_PDU(MyDetailsMessage Message)
        {   // Convert an MyDetailsMessage structure into a byte array
            int iFieldSize1 = MAX_NAME_LENGTH;   // The DataLength is NOT the same as the FieldSize
            int iFieldSize2 = Marshal.SizeOf(Message.iAge);
            int iFieldSize3 = MAX_ADDRESS_LENGTH;       // The DataLength is NOT the same as the FieldSize
            int TotalSize = iFieldSize1 + iFieldSize2 + iFieldSize3;
            byte[] bByteArray = new byte[TotalSize];        // Create a byte array to hold the serialised form of the message
            int iFieldOffset = 0;                           // Create an offset value which ensures the byte array is constructed with correct field separations 
            IntPtr ipWorkingStorage = Marshal.AllocHGlobal(TotalSize);    // Allocate unmanaged memory the same size as the object (ensure adequate working storage)

            int iDataLength1 = Message.cName.Length;   // Get the length of data actually stored in field1 (variable length data, fixed size field)
            for (int iIndex = 0; iIndex < iFieldSize1; iIndex++)
            {   // Byte-by-byte copy data from the data field into the byte array, imposing field size as upper limit for transfer
                if (iIndex < iDataLength1)
                {
                    bByteArray[iIndex + iFieldOffset] = Message.cName[iIndex];
                }
            }

            Marshal.StructureToPtr(Message.iAge, ipWorkingStorage, false);  // Marshal field2 of the object into working storage
            iFieldOffset += iFieldSize1;               // Adjust offset for second field
            Marshal.Copy(ipWorkingStorage, bByteArray, iFieldOffset, iFieldSize2);  // Place the contents of working storage into the byte[] with correct offset
            
            int iDataLength3 = Message.cAddress.Length;  // Get the length of data actually stored in field3 (variable length data, fixed size field)
            iFieldOffset += iFieldSize2;                    // Adjust offset for third field
            for (int iIndex = 0; iIndex < iFieldSize3; iIndex++)
            {   // Byte-by-byte copy data from the data field into the byte array, imposing field size as upper limit for transfer
                if (iIndex < iDataLength3)
                {
                    bByteArray[iIndex + iFieldOffset] = Message.cAddress[iIndex];
                }
            }
            Marshal.FreeHGlobal(ipWorkingStorage);          // Free unmanaged memory
            return bByteArray;
        }

        void SendMessage(MyDetailsMessage Message)
        {
            byte[] bBuffer = Serialize_Message_PDU(Message);
            int iBytesSent;
                try
                {
                    iBytesSent = m_IPC_SendSOCKET.SendTo(bBuffer, m_SendEndPoint);
                    if (0 < iBytesSent)
                    {
                        Console.WriteLine("\nMessage sent   Name: {0}   Address: {1}   Age: {2} ",
                           ByteArrayToString(Message.cName), ByteArrayToString(Message.cAddress), Message.iAge);
                    }
                }
                catch // Catch any errors
                {
                    Console.WriteLine("sendto() Failed!");
                    CloseSocketAndExit();
                }
         }

        void CloseSocketAndExit()
        {
            m_IPC_SendSOCKET.Close();
            Console.WriteLine("Socket closed");
            Console.WriteLine("(press a key to continue)");
            Console.ReadKey();	// Wait for user to type a key (to allow them to read the message before the window is closed)
            System.Environment.Exit(1);
        }

        static byte[] StringToByteArray(string sStr)
        {
            byte[] bArray = new byte[sStr.Length * sizeof(char)];
            System.Buffer.BlockCopy(sStr.ToCharArray(), 0, bArray, 0, bArray.Length);
            return bArray;
        }

        static string ByteArrayToString(byte[] bArray)
        {
            char[] cArray = new char[bArray.Length / sizeof(char)];
            System.Buffer.BlockCopy(bArray, 0, cArray, 0, bArray.Length);
            return new string(cArray);
        }
    }
}

