﻿/*	Application		A program which continually requests more memory until it is refused
	File			Program.cs (Memory_Hog) 
	Purpose			A Memory-intensive program which is used to demonstrate attributes of memory management behaviour
	Author			Richard John Anthony
	Date			January 2015
	
	Function		1. Runs in a loop requesting memory, finds allocation limit by doubling request size until a request is refused
					2. Requests and holds the maximum allocation memory amount discovered.
					3. Displays the amount of memory held

    Note, Unlike C and C++, C# automatically manages memory and in particular it has a Garbage Collector which means that memory is
    automatically freed when no longer in use (detected by the fact that the reference count to a particular block of memory has 
    reduced to zero). Hence in most C# programs there is no need to manually allocate and free memory in the way that this program
    does. However, this particular program is used to investigate the way in which the operating system allocates memory to programs
    and thus the manual allocation and deallocation is necessary to ensure the program's behaviour is predictable (avoiding the automatic 
    garbage collection). 
    To achieve manual memory allocation and deallocation, this program uses: 
        Marshal.AllocHGlobal(amount of memory to allocate);             // The equivallent of C malloc
        Marshal.FreeHGlobal(pointer to previously allocated memory);    // The equivallent of C free
 */

using System;
using System.Runtime.InteropServices;

namespace Memory_Hog
{
class Program
    {
        public const Int32 iAllocationIncrement = 1000000; // 1 million bytes

        static void Main(string[] args)
        {
	        IntPtr piMemoryAllocated;
	        Int32 iAllocationCount = 1;
            bool bLimitReached = false;
	        char cIn;
            string sOut;
	        Console.WriteLine("*******************************************************************************");
	        Console.WriteLine("*   WARNING   -  could crash your computer or loose data  -  read carefully   *");
	        Console.WriteLine("*******************************************************************************");
	        Console.WriteLine("* This program has been designed for the exploration of memory allocation     *");
	        Console.WriteLine("* behaviour. Running this program COULD cause problems with your computer -   *");
	        Console.WriteLine("* do not run this program unless you are happy to risk a crash.               *");
	        Console.WriteLine("* If you do decide to continue to run this program, make sure you have saved  *");
	        Console.WriteLine("* all work, closed all files and shutdown any applications which may corrupt  *");
	        Console.WriteLine("* data. This program should not be run on any computer containing valuable    *");
	        Console.WriteLine("* data resources or running important services; if in doubt DO NOT RUN IT.    *");
	        Console.WriteLine("* The results of running multiple copies of this program, for the purposes of *");
	        Console.WriteLine("* exploring the operating system's behaviour under extreme memory demand      *");
	        Console.WriteLine("* conditions are reported in the accompanying book, in the Resource-View      *");
	        Console.WriteLine("* chapter, so you can read the results and accompanying explanation without   *");
	        Console.WriteLine("* having to actually run this program.                                        *");
	        Console.WriteLine("*******************************************************************************");
	        Console.WriteLine("*                        To continue with this program                        *");
	        Console.WriteLine("*      (which involves dynamic allocation of very large blocks of memory)     *");
            Console.WriteLine("*                          press 'M' followed by '@'                          *");
	        Console.WriteLine("*                                                                             *");
	        Console.WriteLine("*         To exit now without performing any dynamic memory allocation        *");
	        Console.WriteLine("*                            press any other key                              *");
	        Console.WriteLine("*******************************************************************************");
	
	        cIn = (char) Console.Read();
	        if('M' != cIn)
	        {
		        Console.WriteLine("Exiting safely");
		        System.Environment.Exit(1);
	        }
	        cIn = (char) Console.Read();
	        if('@' != cIn)
	        {
		        Console.WriteLine("Exiting safely");
		        System.Environment.Exit(1);
	        }
	       
            do
	        {
                sOut = string.Format("Requesting {0} * {1} bytes (100 megabyte increments)", iAllocationCount, iAllocationIncrement);
                Console.WriteLine(sOut);

		       try
               {
                    piMemoryAllocated = Marshal.AllocHGlobal(iAllocationIncrement * iAllocationCount); // The equivallent of C malloc
                    Marshal.FreeHGlobal(piMemoryAllocated); // The equivallent of C free
			        Console.WriteLine(" ALLOCATED (now freeing)");
			        iAllocationCount *= 2;
		        }
                catch(OutOfMemoryException ex1) // There is insufficient memory to satisfy the request.
		        {	
			        Console.WriteLine(" REFUSED");
			        iAllocationCount /= 2;
                    bLimitReached = true;
                    try
                    {
                        piMemoryAllocated = Marshal.AllocHGlobal(iAllocationIncrement * iAllocationCount); // The equivallent of C malloc
                        sOut = string.Format("{0} * {1} bytes re-allocated and held", iAllocationCount, iAllocationIncrement);
                        Console.WriteLine(sOut);
                    }
                    catch (OutOfMemoryException ex2) // There is insufficient memory to satisfy the request.
                    {
                        Console.WriteLine("Re-allocation attempt failed - No memory held");
                    }
			        break;
		        }
            } while (false == bLimitReached);

	        while(true)
	        {
		        // Wait here to allow exploration, e.g. with Task Manager
	        }
        }
    }
}

